#include "PLCProcess.h"

PLCProcess::PLCProcess()
{
    plc_modbus::CPLCModbus plc_ = new plc_modbus::CPLCModbus();
    plc_process_thread_ = new boost::thread(boost::bind(PLCProcess::thread_func_process, this));
}

PLCProcess::~PLCProcess()
{
    plc_process_thread_->interrupt();
    plc_process_thread_->join();
    delete plc_process_thread_;
}

void PLCProcess::Update(std::vector<std::vector<tf::Transform>> trajectorys)
{
    split_trajectorys(trajectorys);
    trajectory_.bUpdate = true;
}

bool PLCProcess::split_trajectorys(std::vector<std::vector<tf::Transform>> trajectorys)
{
    if (trajectorys.size()>=3)
    {
        std::vector<tf::Transform> left = trajectorys[0];
        std::vector<tf::Transform> right = trajectorys[2];
        trajectory_.left_trajectorys_.clear();
        trajectory_.right_trajectorys_.clear();
        int groups = left.size() / POSE_COUNT + 1;
        for (int i = 0; i < groups; i++)
        {
            trajectory_.left_trajectorys_.push_back(std::vector<tf::Transform>());
            for (int j = 0; j < std::min((int)POSE_COUNT, (int)(left.size() - (i * POSE_COUNT))); j++)
            {
                trajectory_.left_trajectorys_[i].push_back(left[i*POSE_COUNT+j]);
            }
        }
        for (int i = 0; i < groups; i++)
        {
            trajectory_.right_trajectorys_.push_back(std::vector<tf::Transform>());
            for (int j = 0; j < std::min((int)POSE_COUNT, (int)(right.size() - (i * POSE_COUNT))); j++)
            {
                trajectory_.right_trajectorys_[i].push_back(right[i * POSE_COUNT + j]);
            }
        }
        return true;
    }else{
        return false;
    }
}

bool PLCProcess::send_trajectory(std::vector<tf::Transform> trajectory_l, std::vector<tf::Transform> trajectory_r)
{
    std::vector<uint16_t> trajectory_combined = std::vector<uint16_t>();
    std::vector<uint16_t> temp = std::vector<uint16_t>();
    uint16_t *arr = new uint16_t(sizeof(uint16_t) * POSE_COUNT*2);
    temp = absolute_convert_to_relative(trajectory_l);
    trajectory_combined.insert(trajectory_combined.end(),temp.begin(),temp.end());
    temp = absolute_convert_to_relative(trajectory_r);
    trajectory_combined.insert(trajectory_combined.end(), temp.begin(), temp.end());
    if(!trajectory_combined.empty())
    {
        memcpy(arr, &trajectory_combined[0], trajectory_combined.size() * sizeof(uint16_t));
    }
    return plc_.update_trajectory(true, arr);
}
//根据tf绝对路径，获得相对路径vector
std::vector<uint16_t> absolute_convert_to_relative(std::vector<tf::Transform> tf_data)
{
    uint16_t distance = 0;
    tf::Vector3 pos0, pos1;
    std::vector<uint16_t> relative_path = std::vector<uint16_t>();
    for (int i = 0; i < tf_data.size() - 1; i++)
    {
        pos0 = tf_data[i].getOrigin();
        pos1 = tf_data[i+1].getOrigin();
        distance = (uint16_t)(1000*sqrt(pow(pos1.x() - pos0.x(), 2.0) + pow(pos1.y() - pos0.y(), 2.0) + pow(pos1.z() - pos0.z(), 2.0)));
        relative_path.push_back(distance);
    }
    for (int i = POSE_COUNT;i>relative_path.size();i--)
    {
        relative_path.push_back((uint16_t)0);
    }
    return relative_path;
}
//获取当前节点
unsigned int PLCProcess::get_node_index()
{
    return (unsigned int)plc_.get_current_stage();
}
//线程自动判断运动状态
void PLCProcess::thread_func_process(void *process)
{
    PLCProcess *my_process = (PLCProcess *)process;
    int trajectory_group_index = 0;
    if (!my_process)
        return;
    while (true)
    {
        //识别到路径更新指令，初始化轮数值
        if (my_process->trajectory_.bUpdate)
        {
            //my_process->mutex_trajectorys_.lock();
            trajectory_group_index = 0;
            //my_process->mutex_trajectorys_.unlock();
        }

        unsigned int current_node_index = my_process->get_node_index();
        //接收到更新指令，或路径使用完毕，需要更新
        if (trajectory_group_index == 0 || current_node_index == 9)
        {
            if (my_process->trajectory_.left_trajectorys_.size() > trajectory_group_index)
            {
                trajectory_group_index++;
                my_process->send_trajectory(my_process->trajectory_.left_trajectorys_[trajectory_group_index], my_process->trajectory_.right_trajectorys_[trajectory_group_index]);
            }
            else
            {
                //无路径或已更新完毕
                if (current_node_index == 9)
                {
                    printf("路径已运动完成");
                }
                else
                {
                    printf("未找到路径");
                }
            }
        }
    }
}