#include "Planning.h"
#include "make_trajectory.h"
#include <math.h>
#include <time.h>

void Planning::tf_callback(const tf2_msgs::TFMessage::ConstPtr& msg,Planning* p)
{
    static int n=0;
    
    if(msg->transforms[0].header.frame_id=="map" && msg->transforms[0].child_frame_id=="laser")
    {
	double x=msg->transforms[0].transform.translation.x;
    double y=msg->transforms[0].transform.translation.y;
    double z=msg->transforms[0].transform.translation.z;
    double rx=msg->transforms[0].transform.rotation.x;
    double ry=msg->transforms[0].transform.rotation.y;
    double rz=msg->transforms[0].transform.rotation.z;
    double rw=msg->transforms[0].transform.rotation.w;
    clock_t t1,t2;
    t1=clock();
    Planning* plan=p;
    tf::Transform target;
    tf::Matrix3x3 rotation;
    rotation.setRPY(0,0,0);
    target.setBasis(rotation);
    target.setOrigin(tf::Vector3(0,0,0));

    tf::Quaternion quat(rx,ry,rz,rw);
    tf::Transform start(quat,tf::Vector3(x,y,z));

    plan->Make_plan(target,start);

    ROS_INFO(" x= %.3f  y=%.3f\n",x,y);
    t2=clock();
    ROS_INFO(" make plan time = %.3f\n",t2-t1);
    }
}



Planning::Planning()
{
    plan_publisher_thread_=new boost::thread(boost::bind(Planning::thread_publish_path,this));
    
}

Planning::~Planning()
{
    plan_publisher_thread_->interrupt();
    plan_publisher_thread_->join();
    delete plan_publisher_thread_;
}

void Planning::thread_publish_path(Planning* plan)
{
    if(!plan)
	return ;
    ros::Rate loop_rate(5);
    while(ros::ok())
    {
	if(plan->path_.poses.size()!=0)
	{
	    plan->mutex_path_.lock();
	    plan->path_publisher_.publish(plan->path_);
        plan->path_publisher_left_.publish(plan->path_left_);
        plan->path_publisher_right_.publish(plan->path_right_);
	    plan->mutex_path_.unlock();
	
	    loop_rate.sleep();
	}
    }
}

bool Planning::Init()
{
   /* std::string service_name = "move_base/make_plan";
    //等待服务空闲，如果已经在运行这个服务，会等到运行结束。
    while (!ros::service::waitForService(service_name, ros::Duration(3.0)))
    {
	ROS_INFO("Waiting for service move_base/make_plan to become available");
    }
    
    make_plan_client_ = nh_.serviceClient<nav_msgs::GetPlan>(service_name, true);
    if (!make_plan_client_) 
    {
	ROS_FATAL("Could not initialize get plan service from %s",
	make_plan_client_.getService().c_str());
	return false;
    }*/
    
    
    tf_sub_ = nh_.subscribe<tf2_msgs::TFMessage>("tf", 1,boost::bind(&Planning::tf_callback,_1,this));

    path_publisher_ = nh_.advertise<nav_msgs::Path>("get_plan_center",1, true);
    path_publisher_left_ = nh_.advertise<nav_msgs::Path>("get_plan_left",1, true);
    path_publisher_right_ = nh_.advertise<nav_msgs::Path>("get_plan_right",1, true);
    return true;
}

bool Planning::Make_plan(double x0,double y0,double x1,double y1)
{
    nav_msgs::GetPlan srv;
    srv.request.start.header.frame_id ="map";
    srv.request.start.pose.position.x = x0;//初始位置x坐标
    srv.request.start.pose.position.y = y0;//初始位置y坐标
    srv.request.start.pose.orientation.w = 1.0;//方向
    srv.request.goal.header.frame_id = "map";
    srv.request.goal.pose.position.x = x1;//终点坐标
    srv.request.goal.pose.position.y = y1;
    srv.request.goal.pose.orientation.w = 1.0;
    srv.request.tolerance = 0.5;//如果不能到达目标，最近可到的约束
    
    
    // Perform the actual path planner call
//执行实际路径规划器
    if (make_plan_client_.call(srv)) 
    {
//srv.response.plan.poses 为保存结果的容器，遍历取出
	if (!srv.response.plan.poses.empty()) 
	{
        
	    ros::Time current_time, last_time;
	    current_time = ros::Time::now();
	    last_time = ros::Time::now();

	    //nav_msgs::Path path;
	    boost::mutex::scoped_lock lock(mutex_path_);
	    path_=nav_msgs::Path();
	    path_.header.stamp=current_time;
	    path_.header.frame_id="map";
	   // path_.poses.clear();
	    int n=-1;
	    forEach(const geometry_msgs::PoseStamped &p, srv.response.plan.poses) 
	    {
		n++;
		if(n!=0&& n!=srv.response.plan.poses.size()-1)
		{
		    
		    if(n%10!=0)
		    continue;
		}
		double x = p.pose.position.x;
	        double y = p.pose.position.y;
	        double th = 0.0;
	        double vx = 0.1;
	        double vy = -0.1;
	        double vth = 0.1;
		geometry_msgs::PoseStamped this_pose_stamped;
		this_pose_stamped.pose.position.x = x;
		this_pose_stamped.pose.position.y = y;

		geometry_msgs::Quaternion goal_quat = tf::createQuaternionMsgFromYaw(th);
		this_pose_stamped.pose.orientation.x = goal_quat.x;
		this_pose_stamped.pose.orientation.y = goal_quat.y;
		this_pose_stamped.pose.orientation.z = goal_quat.z;
		this_pose_stamped.pose.orientation.w = goal_quat.w;

		this_pose_stamped.header.stamp=current_time;
		this_pose_stamped.header.frame_id="map";
		path_.poses.push_back(this_pose_stamped);
		//ROS_INFO("x = %f, y = %f", p.pose.position.x, p.pose.position.y);
            
            
	    }
	    
	    printf("update :%d points\n",path_.poses.size());
	}
	else {
	    ROS_WARN("Got empty plan");
	}
    }
    else {
    ROS_ERROR("Failed to call service %s - is the robot moving?",
    make_plan_client_.getService().c_str());
    }
    
    return true;
}

bool Planning::Make_plan(tf::Transform start,tf::Transform target)
{
    MakeTrajectory trajectory(start,target);
    tf::Transform left=tf::Transform(tf::Quaternion(0,0,0),tf::Vector3(0.0,0.1,0));
    tf::Transform right=tf::Transform(tf::Quaternion(0,0,0),tf::Vector3(0.0,-0.1,0));
    trajectory.SetLRTransform(left,right);
    trajectory.Make();
    
    if(trajectory.Size()>0)
    {
        ros::Time current_time, last_time;
	    current_time = ros::Time::now();
	    last_time = ros::Time::now();

        boost::mutex::scoped_lock lock(mutex_path_);
	    path_=nav_msgs::Path();
	    path_.header.stamp=current_time;
	    path_.header.frame_id="map";

        path_left_=nav_msgs::Path();
	    path_left_.header.stamp=current_time;
	    path_left_.header.frame_id="map";

        path_right_=nav_msgs::Path();
	    path_right_.header.stamp=current_time;
	    path_right_.header.frame_id="map";

	    // path_.poses.clear();
	    for(int i=0;i<trajectory.Size();++i) 
	    {
            tf::Transform pose=trajectory[i][0];
	        double x = pose.getOrigin().getX();
            double y = pose.getOrigin().getY();
            double th = 0.0;
            double vx = 0.1;
            double vy = -0.1;
            double vth = 0.1;

		    geometry_msgs::PoseStamped this_pose_stamped;
		    this_pose_stamped.pose.position.x = x;
		    this_pose_stamped.pose.position.y = y;

	    	geometry_msgs::Quaternion goal_quat = tf::createQuaternionMsgFromYaw(th);
	    	this_pose_stamped.pose.orientation.x = goal_quat.x;
	    	this_pose_stamped.pose.orientation.y = goal_quat.y;
	    	this_pose_stamped.pose.orientation.z = goal_quat.z;
	    	this_pose_stamped.pose.orientation.w = goal_quat.w;

	    	this_pose_stamped.header.stamp=current_time;
	    	this_pose_stamped.header.frame_id="map";
	    	path_left_.poses.push_back(this_pose_stamped);


            pose=trajectory[i][1];
	        x = pose.getOrigin().getX();
            y = pose.getOrigin().getY();

		    this_pose_stamped.pose.position.x = x;
		    this_pose_stamped.pose.position.y = y;

	    	this_pose_stamped.pose.orientation.x = goal_quat.x;
	    	this_pose_stamped.pose.orientation.y = goal_quat.y;
	    	this_pose_stamped.pose.orientation.z = goal_quat.z;
	    	this_pose_stamped.pose.orientation.w = goal_quat.w;

	    	this_pose_stamped.header.stamp=current_time;
	    	this_pose_stamped.header.frame_id="map";
	    	path_.poses.push_back(this_pose_stamped);

            pose=trajectory[i][2];
	        x = pose.getOrigin().getX();
            y = pose.getOrigin().getY();

		    this_pose_stamped.pose.position.x = x;
		    this_pose_stamped.pose.position.y = y;

	    	this_pose_stamped.pose.orientation.x = goal_quat.x;
	    	this_pose_stamped.pose.orientation.y = goal_quat.y;
	    	this_pose_stamped.pose.orientation.z = goal_quat.z;
	    	this_pose_stamped.pose.orientation.w = goal_quat.w;

	    	this_pose_stamped.header.stamp=current_time;
	    	this_pose_stamped.header.frame_id="map";
	    	path_right_.poses.push_back(this_pose_stamped);

	    }    
	    printf("update :%d points\n",path_.poses.size());
    }
	else {
	    ROS_WARN("Got empty plan");
	}
    return true;
}







